from pathlib import Path
import re
from glob import iglob
from collections import defaultdict

import click

from pkgmt.versioner.versionersetup import VersionerSetup
from pkgmt.versioner.util import complete_version_string
from pkgmt.exceptions import ProjectValidationError


class Deprecations:
    def __init__(self, root_dir=None) -> None:
        self.root_dir = root_dir

        versioner = VersionerSetup(project_root=root_dir)
        self.current = complete_version_string(
            versioner.current_version().replace("dev", "")
        )

    def check(self):
        """Check if there are pending deprecations"""
        deprecations = find_deprecations(root_dir=self.root_dir)

        mapping = defaultdict(lambda: [])

        for dep in deprecations:
            mapping[complete_version_string(dep.version)].append(dep)

        if self.current in mapping:
            matches_out = "\n".join(f"- {item}" for item in mapping[self.current])
            raise ProjectValidationError(
                f"Found the following pending deprecations:\n{matches_out}"
            )


class DeprecationItem:
    def __init__(self, body, version, path) -> None:
        self.body = body
        self.version = version
        self.path = str(Path(path))

    def __eq__(self, o: object) -> bool:
        return (
            self.body == o.body
            and self.version == o.version
            and Path(self.path) == Path(o.path)
        )

    def __hash__(self) -> int:
        return hash((self.body, self.version, self.path))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.body!r}, {self.version!r}, {self.path!r})"

    def __str__(self) -> str:
        return f"{self.body!r} at {self.path!r}"


def _find_deprecations_in_text(text):
    """Find and parse `.. deprecated::` directives in text

    Returns
    -------
    bodies : list
        Content of each `.. deprecated::` directive

    versions : list
        Versions parsed from the body of each directive
    """
    # .. deprecated:: {version} [whitespace]
    # [word chars or space] [version descriptor] [word chars or space]
    bodies = re.findall(
        r"\.\. deprecated::\s+[0-9\.]+\s+([\w,.` \n\-]+\ +[0-9\.]+[\w,`. \n\-]*\n{1})",
        text,
    )

    versions = [re.search(r"[0-9\.]{3,5}", body).group() for body in bodies]

    return bodies, versions


def find_deprecations(root_dir=None):
    deprecations = []

    root_dir = root_dir or "."

    for path in iglob(f"{root_dir}/**/*.py", recursive=True):
        bodies, version = _find_deprecations_in_text(Path(path).read_text())

        for body, version in zip(bodies, version):
            deprecations.append(DeprecationItem(body, version, path))

    return deprecations


@click.command()
@click.argument("root")
def _cli(root):
    Deprecations(root).check()


if __name__ == "__main__":
    _cli()



import re


def is_pre_release(version):
    return "a" in version or "b" in version or "rc" in version


def is_major_version(version):
    _, major, minor = complete_version_string(version.replace("dev", "")).split(".")
    return (minor == "0") or (major == "0" and minor == "0")


def _split_prerelease_part(version):
    if is_pre_release(version):
        prerelease_part = re.search(r"(a|b|rc)\d+", version).group(0)
    else:
        prerelease_part = ""

    return version.replace(prerelease_part, ""), prerelease_part


def complete_version_string(version):
    part_version, part_prerelease = _split_prerelease_part(version)
    elements = len(part_version.split("."))

    if elements < 3:
        missing = 3 - elements
        suffix = ".".join(["0"] * missing)
        version = f"{part_version}.{suffix}"

    if not version.endswith(part_prerelease):
        version = version + part_prerelease

    return version



import time
from datetime import datetime
import json
import subprocess
import concurrent.futures
from pathlib import Path
from glob import iglob
import re
from itertools import chain
from collections import defaultdict
from urllib.parse import urlparse

import requests


class Response:
    def __init__(self, url, code, broken) -> None:
        self.url = url
        self.code = code
        self.broken = broken

    def __hash__(self) -> int:
        return hash(self.url)

    def __eq__(self, another) -> bool:
        return self.url == another

    def __repr__(self) -> str:
        return f"({self.code}) {self.url}"


class LinksInFile:
    def __init__(self, *, valid, invalid) -> None:
        self.valid = valid
        self.invalid = invalid

    def __iter__(self):
        for link in self.valid:
            yield link

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.valid!r})"

    def __eq__(self, other: object) -> bool:
        return self.valid == other


def _make_glob_exp(extension):
    return f"**/*.{extension}"


def _find_match(links, broken):
    result = []

    for link in links:
        if link in broken:
            result.append(broken[link])

    return result


def _read_file(file):
    file = Path(file)

    if file.suffix == ".ipynb":
        # we need to load the notebook's content; otherwise special characters
        # are not resolved correctly since they're double escaped in the
        # JSON file. Also, notebooks are always encoded as UTF-8
        nb = json.loads(file.read_text(encoding="utf-8"))
        return "\n".join(["\n".join(cell["source"]) for cell in nb["cells"]])
    else:
        return file.read_text()


def find_broken_in_files(
    extensions, ignore_substrings=None, verbose=False, broken_http_codes=None
):
    """
    Parameters
    ----------
    extensions : list
        A list of extensions to consider when looking for files. Example:
        `["py", "rst", "md", "ipynb"]`

    find_broken_in_files : str or list
        File extensions to check. Can be a str ("md") or a list such as
        `["md", "rst"]`

    verbose : bool, deefault=False
        Prints broken links

    broken_http_codes : list, default=None
        Only consider these HTTP codes as broken links, example: `[404]`.
        By default, it considers all non-200 HTTP codes as broken.
    """
    if isinstance(extensions, str):
        extensions = [extensions]

    mapping = _find_links_in_files(extensions, ignore_substrings=ignore_substrings)

    broken = {
        response.url: response
        for response in _find_broken_links(mapping, broken_http_codes=broken_http_codes)
    }

    if verbose:
        for file, links in mapping.items():
            if links.invalid:
                print(f"*** Found invalid links in {file} ***")
                print("\n".join(links.invalid))

        print("=" * 80)

        for file, links in mapping.items():
            match = _find_match(links, broken)

            if match:
                print(f"*** {file} ***")
                print("\n".join([repr(m) for m in match]))

        print("=" * 80)

    return broken


def _find_links_in_files(extensions, ignore_substrings=None):
    globs = (iglob(_make_glob_exp(ext), recursive=True) for ext in extensions)
    content = dict()

    tracked_by_git, error = _git_tracked_files()

    for file in chain(*globs):
        if error or file in tracked_by_git:
            text = _read_file(file)
            content[file] = _find(text, ignore_substrings=ignore_substrings)

    return content


def _find_broken_links(mapping, broken_http_codes):
    urls = {item for sublist in mapping.values() for item in sublist}

    broken = []

    checker = LinkChecker()

    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        future_to_url = {
            executor.submit(
                checker.check_if_broken, url, broken_http_codes=broken_http_codes
            ): url
            for url in urls
        }

        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            try:
                response = future.result()
            except Exception as exc:
                print("%r generated an exception: %s" % (url, exc))
            else:

                if response.broken:
                    broken.append(response)

    return broken


def _find(text, ignore_substrings=None):
    """Find links in text"""
    ignore_substrings = ignore_substrings or []
    # source: https://www.geeksforgeeks.org/python-check-url-string/
    regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"  # noqa
    url = re.findall(regex, text)

    candidates = [
        x[0]
        for x in url
        if not any(substr in x[0] for substr in ignore_substrings)
        # TODO: we should add this to the regex, and remove it from here
        and x[0].startswith("http")
    ]

    valid, invalid = _split_valid_invalid(candidates)

    return LinksInFile(valid=valid, invalid=invalid)


def _is_invalid(link):
    # these are most likely templates or examples
    if any(char in link for char in {"{", "}", "{{", "}}"}):
        return True

    return False


def _split_valid_invalid(candidates):
    valid, invalid = [], []

    for candidate in candidates:
        if _is_invalid(candidate):
            invalid.append(candidate)
        else:
            valid.append(candidate)

    return valid, invalid


class LinkChecker:
    def __init__(self) -> None:
        self.last_timestamp = defaultdict(lambda: datetime.min)

    def check_if_broken(self, url, broken_http_codes=None):
        """Check if a link is broken"""
        netloc = urlparse(url).netloc
        then = self.last_timestamp[netloc]
        now = datetime.now()
        seconds = (now - then).seconds
        self.last_timestamp[netloc] = now

        if seconds < 1:
            time.sleep(1 - seconds)

        try:
            res = requests.head(url)
            code = res.status_code
            broken = not res.ok
        except requests.exceptions.ConnectionError:
            code = None
            broken = True
        except requests.exceptions.MissingSchema:
            code = None
            broken = True

        if broken_http_codes:
            broken = code in broken_http_codes
        else:
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/405
            if code == 405:
                broken = False

        response = Response(url, code, broken)

        return response


# copied from soopervisor
def _git_tracked_files():
    """
    Returns
    -------
    list or None
        List of tracked files or None if an error happened
    None of str
        None if successfully retrieved tracked files, str if an error happened
    """
    res = subprocess.run(
        ["git", "ls-tree", "-r", "HEAD", "--name-only"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if not res.returncode:
        return set(res.stdout.decode().splitlines()), None
    else:
        return None, res.stderr.decode().strip()



import os
import warnings
from pathlib import Path
from pkgmt.versioner.abstractversioner import AbstractVersioner


class VersionerSetup(AbstractVersioner):
    def find_package(self):
        path_to_src = Path(self.project_root, "src")

        if not path_to_src.is_dir():
            raise NotADirectoryError(
                f"Expected a directory at {str(path_to_src)!r} but it doesn't exist"
            )

        dirs = sorted(
            [
                f
                for f in os.listdir(path_to_src)
                if Path("src", f).is_dir()
                and not f.endswith(".egg-info")
                and Path(f).name != "__pycache__"
            ]
        )

        if len(dirs) != 1:
            warnings.warn(
                "Found more than one dir, " f"choosing the first one: {dirs[0]}"
            )

        package_name = dirs[0]
        PACKAGE = path_to_src / package_name
        return package_name, PACKAGE

    def version_file(self):
        return "__init__.py"



import sys
from pathlib import Path

import click

from pkgmt import links, config, test, changelog, hook as hook_, versioneer
from pkgmt import new as new_


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--only-404",
    is_flag=True,
    default=False,
    help="Only consider 404 code as broken",
)
def check_links(only_404):
    """Check for broken links"""
    broken_http_codes = None if not only_404 else [404]

    cfg = config.load()["check_links"]
    out = links.find_broken_in_files(
        cfg["extensions"],
        cfg.get("ignore_substrings"),
        verbose=True,
        broken_http_codes=broken_http_codes,
    )

    if out:
        sys.exit(1)


@cli.command()
@click.argument("name")
def new(name):
    """Create new package"""
    new_.package(name)


@cli.command()
@click.option(
    "-f", "--file", type=click.Path(dir_okay=False, exists=True), default="README.md"
)
@click.option("-i", "--inplace", is_flag=True, show_default=True, default=False)
def test_md(file, inplace):
    """Run a markdown file"""
    test.markdown(file, inplace=inplace)


@cli.command()
def check():
    """Run general checks in the project"""
    text = Path("CHANGELOG.md").read_text()
    changelog.CHANGELOG(text).check(verbose=True)


@cli.command()
@click.option(
    "--uninstall",
    is_flag=True,
    default=False,
    help="Uninstall hook",
)
def hook(uninstall):
    """Install pre-push hook"""

    if uninstall:
        hook_.uninstall_hook()
        click.echo("hook uninstalled.")
    else:
        hook_.install_hook()


@cli.command()
@click.option(
    "--yes",
    is_flag=True,
    default=False,
    help="Do not ask for confirmation",
)
def version(yes):
    """Create a new package version"""
    versioneer.version(project_root=".", tag=True, yes=yes)


@cli.command()
@click.argument("tag")
@click.option(
    "--production",
    is_flag=True,
    default=False,
    help="Upload to the production PyPI server",
)
def release(tag, production):
    """Release this package from a given tag"""
    versioneer.upload(tag, production)



import io
import re
import ast
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("src/pkgmt/__init__.py", "rb") as f:
    VERSION = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ).read()


REQUIRES = [
    "toml",
    "pyyaml",
    "requests",
    "click",
]

DEV = [
    "pytest",
    "black",
    "flake8",
    "invoke",
    "twine",
    # optional dependency for test module
    "jupytext",
    "nbclient",
    "ipykernel",
]

# to test markdown files
ALL = [
    "nbclient",
    "jupytext",
    "ipykernel",
]

# to run: pkgmt check
CHECK = [
    # for check-project
    "mistune>=3rc",
]

setup(
    name="pkgmt",
    version=VERSION,
    description=None,
    license=None,
    author=None,
    author_email=None,
    url=None,
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    classifiers=[],
    keywords=[],
    install_requires=REQUIRES,
    extras_require={
        "dev": DEV + CHECK,
        "all": ALL,
        "check": CHECK,
    },
    entry_points={
        "console_scripts": ["pkgmt=pkgmt.cli:cli"],
    },
)



import os
from pathlib import Path
from pkgmt.versioner.abstractversioner import AbstractVersioner


class VersionerNonSetup(AbstractVersioner):
    def __init__(self, version_package, project_root="."):
        self.version_package = version_package
        super().__init__(project_root)

    def find_package(self):

        for path, dirs, files in os.walk(self.project_root):
            if self.version_package in dirs:
                PACKAGE = Path(path, self.version_package)

        package_name = self.version_package
        return package_name, PACKAGE

    def version_file(self):
        return "_version.py"



from pathlib import Path
import stat
import click
import shutil


pre_push_hook = """\
#!/usr/bin/env python3

from pathlib import Path
import sys
import subprocess

def find_root():
    current = Path().resolve()

    while not (current / 'pyproject.toml').exists():
        if current == current.parent:
            sys.exit('Could not find project root.')

        current = current.parent

    return str(current)

current = find_root()

print(f'Running flake8 from {current!r}')
res = subprocess.run(['flake8'])

if res.returncode:
    print()
    sys.exit('***flake8 returned errors. Fix and push again.***')
"""


def _delete_hook(path):
    """Delete a git hook at the given path"""
    if path.exists():
        if path.is_file():
            path.unlink()
        else:
            # in the remote case that it's a directory
            shutil.rmtree(path)

    click.echo(f"Deleted hook located at {path}")


def _install_hook(path_to_hook, content):
    """
    Install a git hook script at the given path
    """
    if path_to_hook.exists():
        click.echo("Overwriting existing hook...")

    path_to_hook.write_text(content)
    # make the file executable
    path_to_hook.chmod(path_to_hook.stat().st_mode | stat.S_IEXEC)


def install_hook():
    if not Path(".git").is_dir():
        raise NotADirectoryError(
            "Expected a .git/ directory in the current working "
            "directory. Run this from the repository root directory."
        )

    parent = Path(".git", "hooks")
    parent.mkdir(exist_ok=True)

    _install_hook(parent / "pre-push", pre_push_hook)
    click.echo("Successfully installed pre-push git hook")


def uninstall_hook():
    _delete_hook(Path(".git", "hooks", "pre-push"))
